@@ -58,6 +58,7 @@ def nan_to_zero(input_tensor):
 
 
 def bytes_to_string(array: np.ndarray, encoding: str = "utf-8"):
+    if array is None: return None
     return [transcript.decode(encoding) for transcript in array]
 
 
@@ -72,15 +73,10 @@ def merge_two_last_dims(x):
     return tf.reshape(x, shape=[b, -1, f * c])
 
 
-def get_rnn(rnn_type):
+def get_rnn(rnn_type: str):
     assert rnn_type in ["lstm", "gru", "rnn"]
-
-    if rnn_type == "lstm":
-        return tf.keras.layers.LSTM
-
-    if rnn_type == "gru":
-        return tf.keras.layers.GRU
-
+    if rnn_type.lower() == "lstm": return tf.keras.layers.LSTM
+    if rnn_type.lower() == "gru": return tf.keras.layers.GRU
     return tf.keras.layers.SimpleRNN
 
 
@@ -153,3 +149,7 @@ def log10(x):
     numerator = tf.math.log(x)
     denominator = tf.math.log(tf.constant(10, dtype=numerator.dtype))
     return numerator / denominator
+
+
+def get_reduced_length(length, reduction_factor):
+    return tf.cast(tf.math.ceil(tf.divide(length, tf.cast(reduction_factor, dtype=length.dtype))), dtype=tf.int32)
